(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Merge_Sort
imports "~~/src/HOL/ex/MergeSort"
begin

text {* more efficient version of merge sort which only iterates ones 
   through the list to cut it into two halves *}

fun split_half :: "'a list \<Rightarrow> 'a list \<times> 'a list"
  where "split_half (x # y # zs) = (let (xs,ys) = split_half zs in (x # xs, y # ys))"
  | "split_half [] = ([], [])"
  | "split_half [x] = ([x],[])"

lemma split_half: assumes "split_half xs = (ys,zs)"
  shows "multiset_of xs = multiset_of ys + multiset_of zs"
  using assms
proof (induct xs arbitrary: ys zs rule: split_half.induct)
  case (1 x y zs xs ys)
  obtain xs' ys' where split: "split_half zs = (xs',ys')" by force
  from 1(1)[OF split] 1(2) split show ?case by (auto simp add: ac_simps)
qed auto

lemma split_half_size_right: assumes split: "split_half xs = (ys,zs)"
  shows "length zs = length xs div 2"
  using assms 
proof (induct xs arbitrary: ys zs rule: split_half.induct)
  case (1 x y zs xs ys)
  obtain xs' ys' where split: "split_half zs = (xs',ys')" by force
  from 1(1)[OF split] 1(2) split show ?case by (auto simp add: ac_simps)
qed auto

lemma split_half_size_left: assumes split: "split_half xs = (ys,zs)"
  shows "length ys = (Suc (length xs) div 2)"
  using assms 
proof (induct xs arbitrary: ys zs rule: split_half.induct)
  case (1 x y zs xs ys)
  obtain xs' ys' where split: "split_half zs = (xs',ys')" by force
  from 1(1)[OF split] 1(2) split show ?case by (auto simp add: ac_simps)
qed auto


context linorder
begin
 
function(sequential) merge_sort :: "'a list \<Rightarrow> 'a list"
where
  "merge_sort [] = []"
| "merge_sort [x] = [x]"
| "merge_sort xs = (\<lambda> (ys,zs). merge (merge_sort ys) (merge_sort zs)) (split_half xs)"
  by pat_completeness auto

termination
proof
  show "wf (measure length)" by simp
next
  fix v vb vc x y
  assume "(x,y) = split_half (v # vb # vc)" 
  hence "split_half (v # vb # vc) = (x,y)" by auto
  from split_half_size_left[OF this]
  show "(x, v # vb # vc) \<in> measure length" by auto
next
  fix v vb vc x y
  assume "(x,y) = split_half (v # vb # vc)" 
  hence "split_half (v # vb # vc) = (x,y)" by auto
  from split_half_size_right[OF this]
  show "(y, v # vb # vc) \<in> measure length" by auto
qed

lemma sorted_merge_sort:
  "sorted (merge_sort xs)"
proof (induct xs rule: merge_sort.induct) 
  case (3 v vb vc)
  obtain xs ys where vc: "split_half (v # vb # vc) = (xs,ys)" by force
  note vc' = vc[symmetric]
  show ?case unfolding merge_sort.simps Let_def vc split sorted_merge 
    using 3(1)[OF vc'] 3(2)[OF vc'] by auto
qed auto
  
lemma multiset_of_merge_sort:
  "multiset_of (merge_sort xs) = multiset_of xs"
proof (induct xs rule: merge_sort.induct)
  case (3 v vb vc)
  obtain xs ys where vc: "split_half (v # vb # vc) = (xs,ys)" by force
  note vc' = vc[symmetric]
  show ?case unfolding merge_sort.simps Let_def vc split 
    using 3(1)[OF vc'] 3(2)[OF vc'] 
      split_half[OF vc] by auto
qed auto

lemma set_of_merge_sort[simp]:
  "set (merge_sort xs) = set xs"
  using arg_cong[OF multiset_of_merge_sort[of xs], of set_of] by simp

theorem merge_sort_sort:
  "sort = merge_sort"
  by (rule ext, rule properties_for_sort) (fact multiset_of_merge_sort sorted_merge_sort)+
end


end
